"""Workchains to interface various quantum software with AiiDA"""
# pylint: disable=c-extension-no-member
# pylint:disable=no-member
from aiida.engine import ToContext, WorkChain, if_
from aiida.orm import Bool, Code, Dict, List, SinglefileData, Str, load_group
from aiida.orm.extras import EntityExtras
from aiida_gaussian.calculations import GaussianCalculation
from aiida_shell import launch_shell_job

from aiida_aimall.calculations import AimqbCalculation
from aiida_aimall.data import AimqbParameters
from aiida_aimall.workchains.calcfunctions import (
    create_wfx_from_retrieved,
    validate_file_ext,
    validate_parser,
    validate_shell_code,
)
from aiida_aimall.workchains.input import BaseInputWorkChain


class QMToAIMWorkChain(WorkChain):
    r"""Workchain to link quantum chemistry jobs without plugins to AIMAll

    Attributes:
        shell_metadata (aiida.orm.Dict): Metadata for the shell calculation
        shell_retrieved (aiida.orm.List): List of files to store in database
        shell_input_file (aiida.orm.SinglefileData): Input file for the calculation software
        shell_cmdline (aiida.orm.Str): Command line terminal command
        wfx_filename (aiida.orm.Str): Filename of the wfx file
        aim_code (aiida.orm.Code): Code for AIMQB
        aim_file_ext (aiida.orm.Str):
        aim_parser (aiida.orm.Str): Entry point for parser to use.
        dry_run (aiida.orm.Bool): Whether or not this is a trial run

    Example:
        ::

            # ORCA Example
            from aiida_shell import ShellCode
            from aiida.orm import load_computer, List, SinglefileData, Str, load_code
            QMToAIMWorkchain = WorkflowFactory('aimall.qmtoaim')
            # already have an orca code setup,  you can use the codeblock above to do so
            code = load_code('orca@cedar')
            builder = QMToAIMWorkchain.get_builder()
            builder.shell_code = code
            pre_str = 'module load StdEnv/2020; module load gcc/10.3.0; module load openmpi/4.1.1; module load orca/5.0.4'
            builder.shell_metadata = Dict(
                {
                    'options': {
                        'withmpi': False,
                        # modules for the compute cluster to load
                        'prepend_text': pre_str,
                        'resources': {
                            'num_machines': 1,
                            'num_mpiprocs_per_machine': 4,
                        },
                        'max_memory_kb': int(3200 * 1.25) * 1024,
                        'max_wallclock_seconds': 3600
                    }
                }
            )
            # again, for tutorial, using a string parsed as file in place of providing an input file
            file_string = '! B3LYP def2-SVP Opt AIM PAL4\n*xyz 0 1\nH 0.0 0.0 0.0\nH 0.0 0.0 1.0\n*'
            input_file = SinglefileData(io.BytesIO(file_string.encode()))
            builder.shell_input_file = input_file
            # get the resulting wfx and opt file, the above command creates a file file.txt
            # so we replace the txt with the output extensions we want
            shell_list = List([input_file.filename.replace('txt','wfx'),input_file.filename.replace('txt','opt'),])
            builder.shell_retrieved = shell_list
            builder.shell_cmdline = Str('{file}')
            builder.aim_code = load_code('aimall@localhost')
            builder.aim_params = AimqbParameters({'nproc':2,'naat':2,'atlaprhocps':True})
            submit(builder)

    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "shell_code",
            validator=validate_shell_code,  # pylint:disable=expression-not-assigned
        )
        spec.input("shell_metadata", valid_type=Dict)
        spec.input("shell_retrieved", valid_type=List)
        spec.input("shell_input_file", valid_type=SinglefileData, required=False)
        spec.input("shell_cmdline", valid_type=Str, required=True)
        spec.input("wfx_filename", valid_type=Str, required=False)
        spec.input("aim_code", valid_type=Code, required=True)
        spec.input(
            "aim_file_ext",
            valid_type=Str,
            validator=validate_file_ext,  # pylint:disable=expression-not-assigned
            required=False,
            default=lambda: Str("wfx"),
        )
        spec.input("aim_params", valid_type=AimqbParameters, required=True)
        spec.input(
            "aim_parser",
            valid_type=Str,
            validator=validate_parser,
            required=False,
            default=lambda: Str("aimall.base"),
        )
        spec.input("dry_run", valid_type=Bool, default=lambda: Bool(False))

        spec.output("parameter_dict", valid_type=Dict)
        spec.outline(cls.shell_job, cls.aim, cls.result)

    def shell_job(self):
        """Launch a shell job"""
        if self.inputs.dry_run.value:
            return self.inputs
        _, node = launch_shell_job(
            self.inputs.shell_code,
            arguments=self.inputs.shell_cmdline.value,
            nodes={"file": self.inputs.shell_input_file},
            outputs=self.inputs.shell_retrieved.get_list(),
            submit=True,
            metadata=self.inputs.shell_metadata.get_dict(),
        )
        out_dict = {"qm": node}
        return ToContext(out_dict)

    def aim(self):
        """Launch an AIMQB calculation"""
        builder = AimqbCalculation.get_builder()
        builder.parameters = self.inputs.aim_params

        if "wfx_filename" not in self.inputs:
            wfx_file = (
                self.inputs.shell_input_file.filename.split(".")[0]
                + f"_{self.inputs.aim_file_ext.value}"
            )
        else:
            wfx_file = self.inputs.wfx_filename.value.replace(".", "_")
        builder.file = self.ctx.qm.base.links.get_outgoing().get_node_by_label(wfx_file)
        builder.code = self.inputs.aim_code
        builder.metadata.options.parser_name = self.inputs.aim_parser.value
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 2}
        # future work, enable group
        # num_atoms = len(
        #     self.ctx.prereor_aim.get_outgoing()
        #     .get_node_by_label("rotated_structure")
        #     .value.split("\n")
        # )
        # #  generalize for substrates other than H
        # builder.group_atoms = List([x + 1 for x in range(0, num_atoms) if x != 1])
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        out_dict = {"aim": process_node}
        return ToContext(out_dict)

    def result(self):
        """Put results in output node"""
        self.out(
            "parameter_dict",
            self.ctx.aim.base.links.get_outgoing().get_node_by_label(
                "output_parameters"
            ),
        )


class GaussianToAIMWorkChain(BaseInputWorkChain):
    r"""A workchain to submit a Gaussian calculation and automatically setup an AIMAll calculation on the output

    Attributes:
        gauss_params (aiida.orm.Dict): Parameters for the Gaussian calculation
        aim_params (AimqbParameters): Parameters for the AIMQB calculation
        gauss_code (aiida.orm.Code): Code for Gaussian software
        frag_label (aiida.orm.Str): Label of the fragment
        wfx_group (aiida.orm.Str): Group to put the wfx file in
        gaussian_group (aiida.orm.Str): Group to put the GaussianCalculation in
        aim_code (aiida.orm.Code): Code for AIMQB software
        dry_run (aiida.orm.Bool): Whether the run is a dry run
        wfx_filename (aiida.orm.Str): Name of the wfx file produced in the calculation

    Example:
        .. code-block:: python

            from aiida import load_profile
            from aiida.plugins import WorkflowFactory, DataFactory
            from aiida.orm import Dict, StructureData, load_code
            import io
            import ase.io
            from aiida.engine import submit
            load_profile()
            GaussianToAIMWorkChain = WorkflowFactory('aimall.gausstoaim')
            AimqbParameters = DataFactory('aimall.aimqb')
            gaussian_input = Dict(
                            {
                                "link0_parameters": {
                                    "%chk": "aiida.chk",
                                    "%mem": "3200MB",  # Currently set to use 8000 MB in .sh files
                                    "%nprocshared": 4,
                                },
                                "functional": "wb97xd",
                                "basis_set": "aug-cc-pvtz",
                                "charge": 0,
                                "multiplicity": 1,
                                "route_parameters": {"opt": None, "Output": "WFX"},
                                "input_parameters": {"output.wfx": None},
                            })
            aim_input = AimqbParameters({'nproc':2,'naat':2,'atlaprhocps':True})
            # For tutorial purpose, representing a xyz file as a string, and parsing it to get strcutre data
            f = io.StringIO(
                            "5\n\n C -0.1 2.0 -0.02\nH 0.3 1.0 -0.02\nH 0.3 2.5 0.8\nH 0.3 2.5 -0.9\nH -1.2 2.0 -0.02"
                        )
            struct_data = StructureData(ase=ase.io.read(f, format="xyz"))
            f.close()
            builder = GaussianToAIMWorkChain.get_builder()
            builder.g16_params = gaussian_input
            builder.aim_params = aim_input
            builder.structure = struct_data
            builder.gauss_code = load_code('gaussian@localhost')
            builder.aim_code = load_code('aimall@localhost')
            submit(builder)
    """

    @classmethod
    def define(cls, spec):
        """Define workchain steps"""
        super().define(spec)
        spec.input("gauss_params", valid_type=Dict, required=True)
        spec.input("aim_params", valid_type=AimqbParameters, required=True)
        spec.input("gauss_code", valid_type=Code)
        spec.input(
            "frag_label",
            valid_type=Str,
            help="Label for substituent fragment, stored as extra",
            required=False,
        )
        spec.input("wfx_group", valid_type=Str, required=False)
        spec.input("gaussian_group", valid_type=Str, required=False)
        spec.input("aim_code", valid_type=Code)
        spec.input("dry_run", valid_type=Bool, default=lambda: Bool(False))
        spec.input("wfx_filename", valid_type=Str, default=lambda: Str("output.wfx"))
        spec.output("parameter_dict", valid_type=Dict)
        spec.outline(
            cls.validate_input,
            if_(cls.is_xyz_input)(cls.create_structure_from_xyz),
            if_(cls.is_smiles_input)(
                cls.get_molecule_inputs_step, cls.string_to_StructureData
            ),
            if_(cls.is_structure_input)(cls.structure_in_context),
            cls.gauss,
            cls.classify_wfx,
            cls.aim,
            cls.result,
        )

    def gauss(self):
        """Run Gaussian calculation"""
        builder = GaussianCalculation.get_builder()
        builder.structure = self.ctx.structure
        builder.parameters = self.inputs.gauss_params
        builder.code = self.inputs.gauss_code
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 4}
        builder.metadata.options.max_memory_kb = int(6400 * 1.25) * 1024
        builder.metadata.options.max_wallclock_seconds = 604800
        builder.metadata.options.additional_retrieve_list = [
            self.inputs.wfx_filename.value
        ]
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        if "gaussian_group" in self.inputs:
            gauss_group = load_group(self.inputs.gaussian_sp_group)
            gauss_group.add_nodes(process_node)
        out_dict = {"gauss": process_node}
        # self.ctx.standard_wfx = process_node.get_outgoing().get_node_by_label("wfx")
        return ToContext(out_dict)

    def classify_wfx(self):
        """Add the wavefunction file from the previous step to the correct group and set the extras"""
        folder_data = self.ctx.gauss.base.links.get_outgoing().get_node_by_label(
            "retrieved"
        )
        self.ctx.wfx = create_wfx_from_retrieved(self.inputs.wfx_filename, folder_data)
        # later scan input parameters for filename

        if "wfx_group" in self.inputs:
            wf_group = load_group(self.inputs.wfx_group)
            wf_group.add_nodes(self.ctx.wfx)
        if "frag_label" in self.inputs:
            struct_extras = EntityExtras(self.ctx.wfx)
            struct_extras.set("smiles", self.inputs.frag_label.value)

    def aim(self):
        """Run Final AIM Calculation"""
        builder = AimqbCalculation.get_builder()
        builder.parameters = self.inputs.aim_params
        builder.file = self.ctx.wfx
        builder.code = self.inputs.aim_code
        # if "frag_label" in self.inputs:
        #     builder.frag_label = self.inputs.frag_label
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 2}
        num_atoms = len(self.ctx.structure.sites)
        #  generalize for substrates other than H
        builder.group_atoms = List([x + 1 for x in range(0, num_atoms) if x != 1])
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        out_dict = {"aim": process_node}
        return ToContext(out_dict)

    def result(self):
        """Put results in output node"""
        self.out(
            "parameter_dict",
            self.ctx.aim.base.links.get_outgoing().get_node_by_label(
                "output_parameters"
            ),
        )


class GenerateWFXToAIMWorkChain(WorkChain):
    r"""Workchain to generate a wfx file from computational chemistry output files and submit that to an AIMQB Calculation

    Attributes:
        input_file (aiida.orm.SinglefileData): File to convert to a wfx file. Should be .cp2k.out or .molden
        aim_params (AimqbParameters): Command line parameters for AIMQB
        aim_code (aiida.orm.Code): AIMQB code

    Example:
        ::

            from aiida import load_profile
            from aiida.plugins import WorkflowFactory, DataFactory
            from aiida.orm import load_node, load_code, SinglefileData
            from aiida.engine import submit

            load_profile()
            GenerateWFXToAIMWorkChain = WorkflowFactory("aimall.wfxtoaim")
            AimqbParameters = DataFactory("aimall.aimqb")
            # predefined Molden input file in database
            single_file = SinglefileData('/absolute/path/to/file')
            aim_params = AimqbParameters({"naat": 2, "nproc": 2, "atlaprhocps": True})
            aim_code = load_code("aimall@localhost")
            builder = GenerateWFXToAIMWorkChain.get_builder()
            builder.input_file = single_file
            builder.aim_params = aim_params
            builder.aim_code = aim_code
            submit(builder)

    Note:
        This workchain uses the IOData module of the Ayer's group Horton to generate the wfx files. Supported file formats
        include .fchk files, molden files (from Molpro, Orca, PSI4, Turbomole, and Molden), and CP2K atom log files. Further
        note that .fchk files can simply be provided directly to an `AimqbCalculation`.

        While IOData accepts other file formats, these formats are the ones available that contain the necessary information
        to generate wfc files
    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input("input_file", valid_type=SinglefileData)
        spec.input("aim_params", valid_type=AimqbParameters)
        spec.input("aim_code")
        spec.output("output_parameters", valid_type=Dict)
        spec.outline(cls.generate_wfx, cls.aim, cls.result)

    def generate_wfx(self):
        """Given SinglefileData generates a wfx file if IOData is capable"""
        _, node = launch_shell_job(
            "iodata-convert",
            arguments="{file} output.wfx",
            nodes={"file": self.inputs.input_file},
            outputs=["output.wfx"],
            submit=True,
        )
        out_dict = {"shell_wfx": node}
        return ToContext(out_dict)

    def aim(self):
        """Run AIM on the generated wfx file"""
        builder = AimqbCalculation.get_builder()
        builder.parameters = self.inputs.aim_params
        builder.file = self.ctx.shell_wfx.base.links.get_outgoing().get_node_by_label(
            "output_wfx"
        )
        builder.code = self.inputs.aim_code
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 2}
        #  generalize for substrates other than H
        process_node = self.submit(builder)
        out_dict = {"aim": process_node}
        return ToContext(out_dict)

    def result(self):
        """Put results in output node"""
        self.out(
            "output_parameters",
            self.ctx.aim.base.links.get_outgoing().get_node_by_label(
                "output_parameters"
            ),
        )



"""Calcfunctions used throughout workchains"""
# pylint: disable=c-extension-no-member
# pylint:disable=no-member
import io
from string import digits

import ase.io
from aiida.engine import calcfunction
from aiida.orm import Dict, SinglefileData, Str, StructureData
from rdkit import Chem
from rdkit.Chem import AllChem, rdmolops, rdqueries
from rdkit.Chem.MolKey.MolKey import BadMoleculeException
from subproptools.sub_reor import rotate_substituent_aiida


@calcfunction
def generate_rotated_structure_aiida(FolderData, atom_dict, cc_dict):
    """Rotates the fragment to the defined coordinate system

    Args:
        FolderData (aiida.orm.FolderData): aim calculation folder
        atom_dict: AIM atom dict
        cc_dict: AIM cc_dict

    Returns:
        Dict with keys 'atom_symbols' and 'geom' containing atomic symbols and the
            the rotated geometry.

    """
    return Dict(rotate_substituent_aiida(FolderData, atom_dict, cc_dict))


def remove_numcharss_from_strlist(in_list):
    """Remove digits from a list of strings. e.g. ['O1','H2','H3'] -> ['O','H','H']

    Args:
        in_list: input list to remove digits from

    Returns:
        output list with the numerical digits removed from each element

    Note:
        The intention for this list is to convert numered atomic symbols, e.g. from Gaussian
            to just symbols

    """
    remove_digits = str.maketrans("", "", digits)
    out_list = [i.translate(remove_digits) for i in in_list]
    return out_list


@calcfunction
def dict_to_structure(fragment_dict):
    """Generate a StructureData for Gaussian inputs

    Args:
        fragment_dict (aiida.orm.Dict): AiiDA orm.Dict with keys 'atom_symbols' and 'geom'

    Returns:
        aiida.orm.StructureData for the molecule

    Note:
        input can be generated, for example, by
            :func:`aiida_aimall.workchains.calcfunctions.generate_rotated_structure_aiida`

    """
    inp_dict = fragment_dict.get_dict()
    symbols = inp_dict["atom_symbols"]
    symbols = remove_numcharss_from_strlist(symbols)
    coords = inp_dict["geom"]
    # outstr is xyz file contents
    outstr = ""
    # numatoms then blank line, then coordinates and symbols for each atom
    outstr += f"{len(symbols)}\n\n"
    for i, symbol in enumerate(symbols):
        if i != len(symbols) - 1:
            outstr = (
                outstr
                + symbol
                + "   "
                + str(coords[i][0])
                + "   "
                + str(coords[i][1])
                + "   "
                + str(coords[i][2])
                + "\n"
            )
        else:
            outstr = (
                outstr
                + symbol
                + "   "
                + str(coords[i][0])
                + "   "
                + str(coords[i][1])
                + "   "
                + str(coords[i][2])
            )
    # create StructureData from .xyz file string
    f = io.StringIO(outstr)
    struct_data = StructureData(ase=ase.io.read(f, format="xyz"))
    f.close()
    return struct_data


def calc_multiplicity(mol):
    """Calculate the multiplicity of a molecule as 2S +1

    Loops over the atoms in the molecule and gets number of radical electrons,
    then converts that number to the multiplicity.

    Args:
        mol: rdkit.Chem molecule object

    Returns:
        integer number of multiplicity

    """
    num_radicals = 0
    for atom in mol.GetAtoms():
        num_radicals += atom.GetNumRadicalElectrons()
    multiplicity = num_radicals + 1
    return multiplicity


def find_attachment_atoms(mol):
    """Given molecule object, find the atoms corresponding to a * and the atom to which that is bound

    Args:
        mol: rdkit molecule object

    Returns:
        molecule with added hydrogens, the * atom object, and the atom object to which that is attached

    Note:
        Assumes that only one * is present in the molecule
    """
    # * has atomic number 0
    query = rdqueries.AtomNumEqualsQueryAtom(0)
    # add hydrogens now
    h_mol_rw = Chem.RWMol(mol)  # Change type of molecule object
    h_mol_rw = Chem.AddHs(h_mol_rw)
    query_ats = h_mol_rw.GetAtomsMatchingQuery(query)
    if len(query_ats) != 1:
        raise ValueError(
            f"Molecule should have one placeholder atom with atomic number 0, found {len(query_ats)}"
        )
    zero_at = query_ats[0]
    # this will be bonded to one atom - whichever atom in the bond is not *, is the one we are looking for
    bond = zero_at.GetBonds()[0]
    begin_atom = bond.GetBeginAtom()
    if begin_atom.GetSymbol() != "*":
        attached_atom = begin_atom
    else:
        attached_atom = bond.GetEndAtom()
    return h_mol_rw, zero_at, attached_atom


def reorder_molecule(h_mol_rw, zero_at, attached_atom):
    """Reindexes the atoms in a molecule, setting attached_atom to index 0, and zero_at to index 1

    Args:
        h_mol_rw: RWMol rdkit object with explicit hydrogens
        zero_at: the placeholder * atom
        attached_atom: the atom bonded to *

    Returns:
        molecule with reordered indices
    """
    zero_at_idx = zero_at.GetIdx()
    zero_at.SetAtomicNum(1)

    attached_atom_idx = attached_atom.GetIdx()
    # Initialize the new index so that our desired atoms are at the indices we want
    first_two_atoms = [attached_atom_idx, zero_at_idx]
    # Add the rest of the indices in original order
    remaining_idx = [
        atom.GetIdx()
        for atom in h_mol_rw.GetAtoms()
        if atom.GetIdx() not in first_two_atoms
    ]
    out_atom_order = first_two_atoms + remaining_idx
    reorder_mol = rdmolops.RenumberAtoms(h_mol_rw, out_atom_order)
    return reorder_mol


def get_xyz(reorder_mol):
    """MMFF optimize the molecule to generate xyz coordiantes

    Args:
        reorder_mol: rdkit.Chem molecule output, output of :func:`aiida_aimall.workchains.calcfunctions.reorder_molecule`

    Returns:
        string of the geometry block of an .xyz file

    """
    AllChem.EmbedMolecule(reorder_mol)
    # not_optimized will be 0 if done, 1 if more steps needed
    max_iters = 200
    for i in range(0, 6):
        not_optimized = AllChem.MMFFOptimizeMolecule(
            reorder_mol, maxIters=max_iters
        )  # Optimize with MMFF94
        # -1 is returned for molecules where there are no heavy atom-heavy atom bonds
        # for these, hopefully the embed geometry is good enough
        # 0 is returned on successful opt
        if not_optimized in [0, -1]:
            break
        if i == 5:
            return "Could not determine xyz coordinates"
        max_iters = max_iters + 200
    xyz_block = AllChem.rdmolfiles.MolToXYZBlock(
        reorder_mol
    )  # pylint:disable=no-member  # Store xyz coordinates
    split_xyz_block = xyz_block.split("\n")
    # first two lines are: number of atoms and blank. Last line is blank
    xyz_lines = split_xyz_block[2 : len(split_xyz_block) - 1]
    xyz_string = "\n".join([str(item) for item in xyz_lines])
    return xyz_string


@calcfunction
def get_substituent_input(smiles: str) -> dict:
    """For a given smiles, determine xyz structure, charge, and multiplicity

    Args:
        smiles (str): SMILEs of substituent to run

    Returns:
        Dict with keys xyz, charge, multiplicity

    Raise:
        ValueError: if molecule cannot be built from SMILES

    """
    mol = Chem.MolFromSmiles(smiles.value)
    # If the mol could not be built, mol will be None
    if not mol:
        raise ValueError(
            f"Molecule could not be constructed for substituent input SMILES {smiles.value}"
        )
    # add hydrogens to the molecule, and find the atom to put at origin and the atom attached to it
    h_mol_rw, zero_at, attached_atom = find_attachment_atoms(mol)
    # Set zero_at to be the first atom, and attached atom as the second, by number
    reorder_mol = reorder_molecule(h_mol_rw, zero_at, attached_atom)
    xyz_string = get_xyz(reorder_mol)
    if xyz_string == "Could not determine xyz coordinates":
        raise BadMoleculeException(
            "Maximum iterations exceeded, could not determine xyz coordinates for f{smiles.value}"
        )
    # get store charge, multiplicity and geometry
    reorder_mol.UpdatePropertyCache()
    charge = Chem.GetFormalCharge(h_mol_rw)
    multiplicity = calc_multiplicity(h_mol_rw)
    out_dict = Dict({"xyz": xyz_string, "charge": charge, "multiplicity": multiplicity})
    return out_dict


@calcfunction
def generate_structure_data(smiles_dict):
    """Take an input xyz string and convert it to StructureData

    Args:
        smiles_dict: output of :func:`aiida_aimall.workchains.calcfunctions.get_substituent_input`

    Returns:
        StructureData of the molecule

    """
    structure_Str = smiles_dict["xyz"]
    structure_str = structure_Str
    # Use the geometry string and create the xyz string for a full .xyz file
    num_atoms = len(structure_str.split("\n"))
    xyz_string = f"{num_atoms}\n\n" + structure_str
    # Convert string to StructureData by encoding it as a file
    f = io.StringIO(xyz_string)
    struct_data = StructureData(ase=ase.io.read(f, format="xyz"))
    f.close()
    return struct_data


@calcfunction
def parameters_with_cm(parameters, smiles_dict):
    """Add charge and multiplicity keys to Gaussian Input

    Args:
        parameters: dictionary to be provided to GaussianCalculation
        smiles_dict: `aiida_aimall.workchains.calcfunctions.get_substituent_input`

    Returns:
        Dict of Gaussian parameters updated with charge and multiplicity

    """
    parameters_dict = parameters.get_dict()
    smiles_dict_dict = smiles_dict.get_dict()
    parameters_dict["charge"] = smiles_dict_dict["charge"]
    parameters_dict["multiplicity"] = smiles_dict_dict["multiplicity"]
    return Dict(parameters_dict)


@calcfunction
def get_wfxname_from_gaussianinputs(gaussian_parameters):
    """Find the .wfx filename from gaussian_parameters

    Check if input parameters was provided to gaussian_parameters, and if so, look for
    .wfx file names supplied. If it was, return the first .wfx filename found

    Args:
        gaussian_parameters: input dictionary to be provided to GaussianCalculation

    Returns:
        Str of .wfx filename

    """
    gaussian_dict = gaussian_parameters.get_dict()
    if "input_parameters" not in gaussian_dict:
        return Str("")
    object_names = list(gaussian_dict["input_parameters"].keys())
    wfx_files = [x for x in object_names if "wfx" in x]
    if len(wfx_files) >= 1:
        return Str(wfx_files[0])
    if len(wfx_files) == 0:
        wfn_files = [x for x in object_names if "wfn" in x]
        if len(wfn_files) >= 1:
            return Str(wfn_files[0])
        if len(wfn_files) == 0:
            return Str("")
    # should not get here
    return Str("")


@calcfunction
def create_wfx_from_retrieved(wfxname, retrieved_folder):
    """Create wavefunction SinglefileData from retrieved folder

    Args:
        wfxname: Str of the name of a .wfx file to get from the retrieved folder
        retrieved_folder: FolderData of a completed GaussianCalculation

    Returns:
        SinglefileData of the .wfx file to find in the FolderData

    """
    wfx_file_string = retrieved_folder.get_object_content(wfxname.value.strip())
    return SinglefileData(io.BytesIO(wfx_file_string.encode()))


def validate_shell_code(node, _):
    """Validate the shell code, ensuring that it is ShellCode or Str

    Args:
        node: input node to check the type for ShellCode or Str

    Returns:
        None if the type is ShellCode or Str, or error string if node is not

    """
    if node.node_type not in [
        "data.core.code.installed.shell.ShellCode.",
        "data.core.str.Str.",
    ]:
        return "the `shell_code` input must be either ShellCode or Str of the command."
    return None


def validate_parser(node, _):
    """Validate the parser, ensuring that the provided value is one of the accepted values.

    Args:
        node: input node to check the type for ShellCode or Str

    Returns:
        None if the value is aimall.base or aimall.group, or an error string if it is not
    """

    if node.value not in ["aimall.base", "aimall.group"]:
        return "the `aim_parser` input must be either aimall.base or aimall.group"
    return None


def validate_file_ext(node, _):
    """Validates that the file extension provided for AIM is wfx, wfn or fchk

    Args:
        node: node to check the value of to ensure it is in a supported format

    Returns:
        None if the type is ShellCode or Str, or error string if node is not

    """
    if node.value not in ["wfx", "wfn", "fchk"]:
        return "the `aim_file_ext` input must be a valid file format for AIMQB: wfx, wfn, or fchk"
    return None


@calcfunction
def get_molecule_str_from_smiles(smiles):
    """For a given smiles, determine xyz structure, charge, and multiplicity

    Args:
        smiles: SMILEs of substituent to run

    Returns:
        Dict with keys xyz, charge, multiplicity

    """
    mol = Chem.MolFromSmiles(smiles.value)
    if not mol:
        raise ValueError(
            f"Molecule could not be constructed for substituent input SMILES {smiles.value}"
        )
    h_mol_rw = Chem.RWMol(mol)  # Change type of molecule object
    h_mol_rw = Chem.AddHs(h_mol_rw)
    xyz_string = get_xyz(h_mol_rw)
    if xyz_string == "Could not determine xyz coordinates":
        raise BadMoleculeException(
            "Maximum iterations exceeded, could not determine xyz coordinates for f{smiles.value}"
        )
    h_mol_rw.UpdatePropertyCache()
    charge = Chem.GetFormalCharge(h_mol_rw)
    multiplicity = calc_multiplicity(h_mol_rw)
    out_dict = Dict({"xyz": xyz_string, "charge": charge, "multiplicity": multiplicity})
    return out_dict


@calcfunction
def xyzfile_to_StructureData(xyz_SFD):
    """Convert the xyz file provided as SinglefileData to StructureData"""
    with xyz_SFD.as_path() as filepath:
        return StructureData(ase=ase.io.read(filepath, format="xyz"))



"""Subclasses of `FromGroupSubmissionController` designed to prevent too many running processes

The entire :func:`aiida_aimall.workchains.subparam.SubstituentParameterWorkChain` can be replicated
by linking these together.

Provides controllers for the `AimReorWorkChain`, `AimqbCalculations`, `GaussianCalculation`
and `SmilesToGaussianWorkChain`.

"""
# pylint:disable=too-many-positional-arguments
from aiida import orm
from aiida.orm import Dict, Int, Str
from aiida.plugins import CalculationFactory, WorkflowFactory
from aiida_submission_controller import FromGroupSubmissionController

from aiida_aimall.data import AimqbParameters


class SmilesToGaussianController(FromGroupSubmissionController):
    """A controller for submitting :func:`aiida_aimall.workchains.param_parts.SmilesToGaussianWorkChain`

    Args:
        parent_group_label (`str`): group label which contains various SMILES as orm.Str nodes
        group_label (`str`): group in which to put the GaussianCalculations
        wfxgroup (`str`): group in which to store the resulting wfx files
        max_concurrent (`int`): maximum number of concurrent processes
        code_label (`str`): label of Gaussian code, e.g. gaussian@cedar
        gauss_opt_params(`Dict`): Dict of Gaussian parameters to use
        nprocs (`int`): number of processors for gaussian calculation
        mem_mb (`int`): amount of memory in MB for Gaussian calculation
        time_s (`int`): wallclock time in seconds for Gaussian calculation

    Returns:
        Controller object, periodically use run_in_batches to submit new results

    Example:
        In a typical use case of controllers, it is beneficial to check for new jobs periodically to submit.
            Either there may be new members of the parent_group to run, or some of the currently running jobs have run.
            So once a controller is defined, we can run it in a loop.

        ::

            controller = SmilesToGaussianController(
                code_label='gaussian@localhost',
                parent_group_label = 'input_smiles', # Add structures to run to input_smiles group
                group_label = 'gaussianopt', # Resulting nodes will be in the gaussianopt group
                max_concurrent = 1,
                wfxgroup = "opt_wfx",
                gauss_opt_params = Dict(dict={
                    'link0_parameters': {
                        '%chk':'aiida.chk',
                        "%mem": "4000MB",
                        "%nprocshared": 4,
                    },
                    'functional':'wb97xd',
                    'basis_set':'aug-cc-pvtz',
                    'route_parameters': { 'opt':None, 'freq':None},
                    }),
                nprocs = 4,
                mem_mb = 6400,
                time_s = 24*3600*7
                )

            while True:
                #submit Gaussian batches every hour
                controller.submit_new_batch()
                time.sleep(3600)

    """

    # pylint:disable=too-many-arguments
    parent_group_label: str
    group_label: str
    code_label: str
    max_concurrent: int
    gauss_opt_params: dict
    wfxgroup: str
    nprocs: int
    mem_mb: int
    time_s: int

    WORKFLOW_ENTRY_POINT: str = "aimall.smitogauss"

    def __init__(
        self,
        code_label: str,
        gauss_opt_params: dict,
        wfxgroup: str,
        nprocs: int,
        mem_mb: int,
        time_s: int,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.code_label = code_label
        self.gauss_opt_params = gauss_opt_params
        self.wfxgroup = wfxgroup
        self.nprocs = nprocs
        self.mem_mb = mem_mb
        self.time_s = time_s

    def get_extra_unique_keys(self):
        """Returns a tuple of extras keys in the order needed"""
        return ("smiles",)

    def get_inputs_and_processclass_from_extras(self, extras_values):
        """Constructs input for a GaussianCalculation from extra_values"""
        code = orm.load_code(self.code_label)
        smiles = self.get_parent_node_from_extras(extras_values)
        inputs = {
            "smiles": smiles,
            "gaussian_code": code,
            "gaussian_parameters": Dict(self.gauss_opt_params),
            "wfxgroup": Str(self.wfxgroup),
            "nprocs": Int(self.nprocs),
            "mem_mb": Int(self.mem_mb),
            "time_s": Int(self.time_s),
        }
        return inputs, WorkflowFactory(self.WORKFLOW_ENTRY_POINT)


class AimReorSubmissionController(FromGroupSubmissionController):
    """A controller for submitting :func:`aiida_aimall.workchains.param_parts.AIMAllReorWorkChain`.

    Args:
        parent_group_label: the string of a group label which contains various structures as orm.Str nodes
        group_label: the string of the group to put the GaussianCalculations in
        max_concurrent: maximum number of concurrent processes.
        code_label: label of code, e.g. gaussian@cedar
        reor_group: group in which to place the reoriented structures.
        aimparameters: dict of parameters for running AimQB, to be converted to AimqbParameters by the controller

    Returns:
        Controller object, periodically use run_in_batches to submit new results

    Note:
        A typical use case is using this as a controller on wfx files created by GaussianCalculation. In that case,
            match the `parent_group_label` here to the `wfxgroup` provided to the GaussianCalculation.
            In GaussianOptWorkchain, this is `opt_wfx` by default

    Example:
        In a typical use case of controllers, it is beneficial to check for new jobs periodically to submit.
            Either there may be new members of the parent_group to run, or some of the currently running jobs have run.
            So once a controller is defined, we can run it in a loop.

        ::

            controller = AimReorSubmissionController(
                code_label='aimall@localhost',
                parent_group_label = 'wfx', # Add wfx files to run to group wfx
                group_label = 'aim',
                max_concurrent = 1,
                reor_group = "reor_str"
                aimparameters = {"naat": 2, "nproc": 2, "atlaprhocps": True}
            )

            while True:
                #submit AIM batches every 5 minutes
                i = i+1
                controller.submit_new_batch()
                time.sleep(300)

    """

    parent_group_label: str
    group_label: str
    max_concurrent: int
    code_label: str
    reor_group: str
    aimparameters: dict

    WORKFLOW_ENTRY_POINT: str = "aimall.aimreor"

    def __init__(
        self,
        code_label: str,
        reor_group: str,
        aimparameters,
        *args,
        **kwargs,
    ):
        """initialize class"""
        super().__init__(*args, **kwargs)
        self.code_label = code_label
        self.reor_group = reor_group
        self.aimparameters = aimparameters

    # @validator("code_label")
    # # def _check_code_plugin(self, value):
    # #     """validate provided code label:
    # #     Note: unsure if works
    # #     """
    # #     plugin_type = orm.load_code(value).default_calc_job_plugin
    # #     if plugin_type == "aiida_aimall.calculations:AimqbCalculation":
    # #         return value
    # #     raise ValueError(
    # #         f"Code with label `{value}` has incorrect plugin type: `{plugin_type}`"
    # #     )

    def get_extra_unique_keys(self):
        """Returns a tuple of extras keys in the order needed"""
        return ("smiles",)

    def get_inputs_and_processclass_from_extras(self, extras_values):
        """Constructs input for a :func:`aiida_aimall.workchains.param_parts.AIMAllReorWorkChain` from extra_values"""
        code = orm.load_code(self.code_label)
        # AimqbParameters = DataFactory("aimall")

        inputs = {
            "aim_code": code,
            "aim_params": AimqbParameters(parameter_dict=self.aimparameters),
            "file": self.get_parent_node_from_extras(extras_values),
            "frag_label": Str(extras_values[0]),
            "reor_group": Str(self.reor_group),
        }
        return inputs, WorkflowFactory(self.WORKFLOW_ENTRY_POINT)


class AimAllSubmissionController(FromGroupSubmissionController):
    """A controller for submitting :func:`aiida_aimall.calculations.AimqbCalculation`.

    Args:
        parent_group_label: the string of a group label which contains various structures as orm.Str nodes
        group_label: the string of the group to put the GaussianCalculations in
        max_concurrent: maximum number of concurrent processes. Expected behaviour is to set to a large number
            since we will be submitting to Cedar which will manage
        code_label: label of code, e.g. gaussian@cedar
        aimparameters: dict of parameters for running AimQB, to be converted to
            :func:`aiida_aimall.data.AimqbParameters` by the controller
        aimparser: str of which AIM parser to use: aimall.base for `AimqbBaseParser` or
            aimall.group for `AimqbGroupParser`

    Returns:
        Controller object, periodically use run_in_batches to submit new results

    Note:
        A typical use case is using this as a controller on wfx files created by `GaussianCalculation`. In that case,
            match the `parent_group_label` here to the `wfxgroup` provided to the `GaussianCalculation`.
            In GaussianSubmissionController, this is `reor_wfx`

    Example:
        In a typical use case of controllers, it is beneficial to check for new jobs periodically to submit.
            Either there may be new members of the parent_group to run, or some of the currently running jobs have run.
            So once a controller is defined, we can run it in a loop.

        ::

            controller = AimAllSubmissionController(
                code_label='aimall@localhost',
                parent_group_label = 'wfx', # Add wfx files to run to group wfx
                group_label = 'aim_reor',
                max_concurrent = 1,
                aim_parser = 'aimqb.group'
                aimparameters = {"naat": 2, "nproc": 2, "atlaprhocps": True}
            )

            while True:
                #submit AIM batches every 5 minutes
                i = i+1
                controller.submit_new_batch()
                time.sleep(300)

    """

    parent_group_label: str
    group_label: str
    max_concurrent: int
    code_label: str
    aim_parser: str
    aimparameters: dict

    CALCULATION_ENTRY_POINT: str = "aimall.aimqb"

    def __init__(
        self,
        code_label: str,
        aim_parser: str,
        aimparameters: dict,
        *args,
        **kwargs,
    ):
        """Initialize the class, modifying with new values"""
        super().__init__(*args, **kwargs)
        self.code_label = code_label
        self.aim_parser = aim_parser
        self.aimparameters = aimparameters

    # @validator("code_label")
    # def _check_code_plugin(self, value):
    #     """Make sure code label works.

    #     Note: unsure this works"""
    #     plugin_type = orm.load_code(value).default_calc_job_plugin
    #     if plugin_type == "aiida_aimall.calculations:AimqbCalculation":
    #         return value
    #     raise ValueError(
    #         f"Code with label `{value}` has incorrect plugin type: `{plugin_type}`"
    #     )

    def get_extra_unique_keys(self):
        """Returns a tuple of extras keys in the order needed"""
        return ("smiles",)

    def get_inputs_and_processclass_from_extras(self, extras_values):
        """Constructs input for a AimQBCalculation from extra_values"""
        code = orm.load_code(self.code_label)
        # AimqbParameters = DataFactory("aimall")
        inputs = {
            "code": code,
            # "frag_label": Str(extras_values[0]),
            "parameters": AimqbParameters(parameter_dict=self.aimparameters),
            "file": self.get_parent_node_from_extras(extras_values),
            "metadata": {
                "options": {
                    "resources": {"num_machines": 1, "num_mpiprocs_per_machine": 2},
                    "parser_name": self.aim_parser,
                }
            },
        }
        return inputs, CalculationFactory(self.CALCULATION_ENTRY_POINT)


class GaussianSubmissionController(FromGroupSubmissionController):
    """A controller for submitting `GaussianCalculation`.

    Args:
        parent_group_label: the string of a group label which contains various structures as orm.Str nodes
        group_label: the string of the group to put the GaussianCalculations in
        max_concurrent: maximum number of concurrent processes. Expected behaviour is to set to a large number
            since we will be submitting to Cedar which will manage
        code_label: label of code, e.g. gaussian@cedar
        gauss_sp_params: dictionary of parameters to use in gaussian calculation
        wfxname: Name of the wfx file


    Returns:
        Controller object, periodically use run_in_batches to submit new results

    Note:
        A typical use case is using this as a controller on Str structures generated by AIMAllReor workchain. These are by
            default assigned to the `reor_structs` group, so have `parent_group_label` match that

    Note:
        In overall workchain(fragment->optimize->aim+rotate->single point->aim), this is the single point step.
        Process continues and finishes in AimAllSubmissionController

    Example:
        In a typical use case of controllers, it is beneficial to check for new jobs periodically to submit.
            Either there may be new members of the parent_group to run, or some of the currently running jobs have run.
            So once a controller is defined, we can run it in a loop.

        ::

            controller = GaussianSubmissionController(
                code_label='gaussian@localhost',
                parent_group_label = 'struct', # Add structures to run to struct group
                group_label = 'gaussiansp', # Resulting nodes will be in the gaussiansp group
                max_concurrent = 1,
                wfxname='output.wfx'
                gauss_sp_params = Dict(dict={
                    'link0_parameters': {
                        '%chk':'aiida.chk',
                        "%mem": "4000MB",
                        "%nprocshared": 4,
                    },
                    'functional':'wb97xd',
                    'basis_set':'aug-cc-pvtz',
                    'charge': 0,
                    'multiplicity': 1,
                    'route_parameters': {'nosymmetry':None, 'Output':'WFX'},
                    "input_parameters": {"output.wfx": None},
                    })
            )

            while True:
                #submit Gaussian batches every hour
                controller.submit_new_batch()
                time.sleep(3600)

    """

    parent_group_label: str
    group_label: str
    max_concurrent: int
    code_label: str
    gauss_sp_params: dict
    wfxname: str
    # GaussianWFXCalculation entry point as defined in aiida-aimall pyproject.toml
    CALCULATION_ENTRY_POINT: str = "gaussian"

    def __init__(
        self,
        code_label: str,
        gauss_sp_params: dict,
        wfxname: str,
        *args,
        **kwargs,
    ):
        """Initialize the class, modifying with new values"""
        super().__init__(*args, **kwargs)
        self.code_label = code_label
        self.gauss_sp_params = gauss_sp_params
        self.wfxname = wfxname

    # @validator("code_label")
    # def _check_code_plugin(self, value):
    #     """validate that the code label is a GaussianWFXCalculation

    #     Note: unsure if this part works
    #     """
    #     plugin_type = orm.load_code(value).default_calc_job_plugin
    #     if plugin_type == "aiida_aimall.calculations:GaussianWFXCalculation":
    #         return value
    #     raise ValueError(
    #         f"Code with label `{value}` has incorrect plugin type: `{plugin_type}`"
    #     )

    def get_extra_unique_keys(self):
        """Returns a tuple of extras keys in the order needed"""
        return ("smiles",)

    def get_inputs_and_processclass_from_extras(self, extras_values):
        """Constructs input for a GaussianWFXCalculation from extra_values

        Note: adjust the metadata options later for 6400MB and 7days runtime
        """
        code = orm.load_code(self.code_label)
        structure = self.get_parent_node_from_extras(extras_values)
        inputs = {
            "code": code,
            "parameters": Dict(self.gauss_sp_params),
            "structure": structure,
            "metadata": {
                "options": {
                    "resources": {"num_machines": 1, "tot_num_mpiprocs": 1},
                    "max_memory_kb": int(3200 * 1.25) * 1024,
                    "max_wallclock_seconds": 604800,
                    "additional_retrieve_list": [self.wfxname.strip()],
                }
            },
        }
        return inputs, CalculationFactory(self.CALCULATION_ENTRY_POINT)



"""`WorkChain` for calculating substituent parameters developed by authors"""
# pylint: disable=c-extension-no-member
# pylint:disable=no-member
# pylint:disable=too-many-lines
from aiida.engine import ToContext, if_
from aiida.orm import Bool, Code, Dict, List, Str, load_group
from aiida.orm.extras import EntityExtras
from aiida_gaussian.calculations import GaussianCalculation

from aiida_aimall.calculations import AimqbCalculation
from aiida_aimall.data import AimqbParameters
from aiida_aimall.workchains.calcfunctions import (
    create_wfx_from_retrieved,
    get_substituent_input,
)
from aiida_aimall.workchains.input import BaseInputWorkChain
from aiida_aimall.workchains.param_parts import AIMAllReorWorkChain


class SubstituentParameterWorkChain(BaseInputWorkChain):
    r"""A workchain to calculate properties of substituents, R, in R-H molecules.

    This is a multistep calculation, consisting of a Gaussian calculation, an AIMQB calculation,
    Python reorientation to the defined coordinate system, a Gaussian single point calculation,
    and a final AIMQB calculation on the single point wfx calculation. Substituent Properties are
    then extracted using the AimqbGroupParser.

    Attributes:
        gauss_opt_params (aiida.orm.Dict): Parameters for the Gaussian optimization calculations
        gauss_sp_params (aiida.orm.Dict): Parameters for the Gaussian single point calculations
        aim_params (AimqbParameters): Command line parameters for AIMQB
        gauss_code (aiida.orm.Code): Code for Gaussian software
        frag_label (aiida.orm.Str): Optional fragment label for the calculation
        opt_wfx_group (aiida.orm.Str): Optional group to put optimization wavefunctions in
        sp_wfx_group (aiida.orm.Str): Optional group to put single point wavefunctions in
        gaussian_opt_group (aiida.orm.Str): Optional group to put optimization GaussianCalculations in
        gaussian_sp_group (aiida.orm.Str): Optional group to put single point GaussianCalculations in
        wfx_filename (aiida.orm.Str): Optional wfx file name
        aim_code (aiida.orm.Code): Code for AIMQB software
        dry_run (aiida.orm.Bool): Whether or not this is a dry run of the WorkChain

    Note:
        Here, the group for a substiuent is defined in an R-H molecule. Atom 1 is the atom in
        the group R that is attached to the hydrogen, and the hydrogen should be atom 2. These
        align with the default settings of an AimqbCalculation using an AimqbGroupParser.

    Example:
        ::

            from aiida.plugins import WorkflowFactory, DataFactory
            from aiida.orm import Dict, StructureData, load_code
            from aiida.engine import submit
            from aiida import load_profile
            import io
            import ase.io

            load_profile()

            SubstituentParameterWorkchain = WorkflowFactory('aimall.subparam')
            AimqbParameters = DataFactory('aimall.aimqb')
            aim_input = AimqbParameters({'nproc':2,'naat':2,'atlaprhocps':True})
            gaussian_opt = Dict(
                        {
                            "link0_parameters": {
                                "%chk": "aiida.chk",
                                "%mem": "3200MB",  # Currently set to use 8000 MB in .sh files
                                "%nprocshared": 4,
                            },
                            "functional": "wb97xd",
                            "basis_set": "aug-cc-pvtz",
                            "charge": 0,
                            "multiplicity": 1,
                            "route_parameters": {"opt": None, "Output": "WFX"},
                            "input_parameters": {"output.wfx": None},
                        }
            )
            gaussian_sp = Dict(
                        {
                            "link0_parameters": {
                                "%chk": "aiida.chk",
                                "%mem": "3200MB",  # Currently set to use 8000 MB in .sh files
                                "%nprocshared": 4,
                            },
                            "functional": "wb97xd",
                            "basis_set": "aug-cc-pvtz",
                            "charge": 0,
                            "multiplicity": 1,
                            "route_parameters": {"nosymmetry": None, "Output": "WFX"},
                            "input_parameters": {"output.wfx": None},
                        }
            )
            f = io.StringIO(
                            "5\n\n C -0.1 2.0 -0.02\nH 0.3 1.0 -0.02\nH 0.3 2.5 0.8\nH 0.3 2.5 -0.9\nH -1.2 2.0 -0.02"
                        )
            struct_data = StructureData(ase=ase.io.read(f, format="xyz"))
            f.close()
            builder = SubstituentParameterWorkchain.get_builder()
            builder.g16_code = load_code('gaussian@localhost')
            builder.aim_code = load_code('aimall@localhost')
            builder.g16_opt_params = gaussian_opt
            builder.g16_sp_params = gaussian_sp
            builder.structure = struct_data
            builder.aim_params = aim_input
            submit(builder)

    """

    @classmethod
    def define(cls, spec):
        """Define workchain steps"""
        super().define(spec)
        spec.input("gauss_opt_params", valid_type=Dict, required=True)
        spec.input("gauss_sp_params", valid_type=Dict, required=True)
        spec.input("aim_params", valid_type=AimqbParameters, required=True)
        spec.input("gauss_code", valid_type=Code)
        spec.input(
            "frag_label",
            valid_type=Str,
            help="Label for substituent fragment, stored as extra",
            required=False,
        )
        spec.input("opt_wfx_group", valid_type=Str, required=False)
        spec.input("sp_wfx_group", valid_type=Str, required=False)
        spec.input("gaussian_opt_group", valid_type=Str, required=False)
        spec.input("gaussian_sp_group", valid_type=Str, required=False)
        spec.input(
            "wfx_filename",
            valid_type=Str,
            required=False,
            default=lambda: Str("output.wfx"),
        )
        # spec.input("file", valid_type=SinglefileData)
        # spec.output('aim_dict',valid_type=Dict)
        spec.input("aim_code", valid_type=Code)
        spec.input("dry_run", valid_type=Bool, default=lambda: Bool(False))
        # spec.input("frag_label", valid_type=Str)
        # spec.output("rotated_structure", valid_type=Str)
        spec.output("parameter_dict", valid_type=Dict)
        spec.outline(
            cls.validate_input,
            if_(cls.is_xyz_input)(cls.create_structure_from_xyz),
            if_(cls.is_smiles_input)(
                cls.get_substituent_inputs_step, cls.string_to_StructureData
            ),
            if_(cls.is_structure_input)(cls.structure_in_context),
            cls.gauss_opt,
            cls.classify_opt_wfx,
            cls.aim_reor,
            cls.gauss_sp,
            cls.classify_sp_wfx,
            cls.aim,
            cls.result,
        )

    def get_substituent_inputs_step(self):
        """Get a dictionary of the substituent input for a given SMILES"""
        self.ctx.smiles_geom = get_substituent_input(self.inputs.smiles)

    def gauss_opt(self):
        """Submit the Gaussian optimization"""
        builder = GaussianCalculation.get_builder()
        builder.structure = self.ctx.structure
        builder.parameters = self.inputs.gauss_opt_params
        builder.code = self.inputs.gauss_code
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 4}
        builder.metadata.options.max_memory_kb = int(6400 * 1.25) * 1024
        builder.metadata.options.max_wallclock_seconds = 604800
        builder.metadata.options.additional_retrieve_list = [
            self.inputs.wfx_filename.value
        ]
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        if "gaussian_opt_group" in self.inputs:
            gauss_opt_group = load_group(self.inputs.gaussian_opt_group)
            gauss_opt_group.add_nodes(process_node)
        out_dict = {"opt": process_node}
        # self.ctx.standard_wfx = process_node.get_outgoing().get_node_by_label("wfx")
        return ToContext(out_dict)

    def classify_opt_wfx(self):
        """Add the wavefunction file from the previous step to the correct group and set the extras"""
        folder_data = self.ctx.opt.base.links.get_outgoing().get_node_by_label(
            "retrieved"
        )
        # later scan input parameters for filename
        wfx_file = create_wfx_from_retrieved(
            self.inputs.wfx_filename.value, folder_data
        )
        self.ctx.opt_wfx = wfx_file

        if "opt_wfx_group" in self.inputs:
            opt_wf_group = load_group(self.inputs.opt_wfx_group)
            opt_wf_group.add_nodes(wfx_file)
        if "frag_label" in self.inputs:
            struct_extras = EntityExtras(wfx_file)
            struct_extras.set("smiles", self.inputs.frag_label.value)

    def aim_reor(self):
        """Submit the Aimqb calculation and reorientation"""
        builder = AIMAllReorWorkChain.get_builder()
        builder.aim_params = self.inputs.aim_params
        builder.file = self.ctx.opt_wfx
        builder.aim_code = self.inputs.aim_code
        # builder.dry_run = self.inputs.dry_run
        if "frag_label" in self.inputs:
            builder.frag_label = self.inputs.frag_label
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        out_dict = {"prereor_aim": process_node}
        return ToContext(out_dict)

    def gauss_sp(self):
        """Run Gaussian Single Point calculation"""
        builder = GaussianCalculation.get_builder()
        builder.structure = (
            self.ctx.prereor_aim.base.links.get_outgoing().get_node_by_label(
                "rotated_structure"
            )
        )
        builder.parameters = self.inputs.gauss_sp_params
        builder.code = self.inputs.gauss_code
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 4}
        builder.metadata.options.max_memory_kb = int(6400 * 1.25) * 1024
        builder.metadata.options.max_wallclock_seconds = 604800
        builder.metadata.options.additional_retrieve_list = [
            self.inputs.wfx_filename.value
        ]
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        if "gaussian_sp_group" in self.inputs:
            gauss_sp_group = load_group(self.inputs.gaussian_sp_group)
            gauss_sp_group.add_nodes(process_node)
        out_dict = {"sp": process_node}
        # self.ctx.standard_wfx = process_node.get_outgoing().get_node_by_label("wfx")
        return ToContext(out_dict)

    def classify_sp_wfx(self):
        """Add the wavefunction file from the previous step to the correct group and set the extras"""
        folder_data = self.ctx.sp.base.links.get_outgoing().get_node_by_label(
            "retrieved"
        )
        # later scan input parameters for filename
        wfx_file = create_wfx_from_retrieved(
            self.inputs.wfx_filename.value, folder_data
        )
        self.ctx.sp_wfx = wfx_file

        if "sp_wfx_group" in self.inputs:
            sp_wf_group = load_group(self.inputs.sp_wfx_group)
            sp_wf_group.add_nodes(wfx_file)
        if "frag_label" in self.inputs:
            struct_extras = EntityExtras(wfx_file)
            struct_extras.set("smiles", self.inputs.frag_label.value)

    def aim(self):
        """Run Final AIM Calculation"""
        builder = AimqbCalculation.get_builder()
        builder.parameters = self.inputs.aim_params
        builder.file = self.ctx.sp_wfx
        builder.code = self.inputs.aim_code
        builder.metadata.options.parser_name = "aimall.group"
        builder.metadata.options.resources = {"num_machines": 1, "tot_num_mpiprocs": 2}
        num_atoms = len(
            self.ctx.prereor_aim.base.links.get_outgoing()
            .get_node_by_label("rotated_structure")
            .sites
        )
        #  generalize for substrates other than H
        builder.group_atoms = List([x + 1 for x in range(0, num_atoms) if x != 1])
        if self.inputs.dry_run.value:
            return self.inputs
        process_node = self.submit(builder)
        out_dict = {"final_aim": process_node}
        return ToContext(out_dict)

    def result(self):
        """Put results in output node"""
        self.out(
            "parameter_dict",
            self.ctx.final_aim.base.links.get_outgoing().get_node_by_label(
                "output_parameters"
            ),
        )



# pylint:disable=too-many-function-args
"""
Parsers provided by aiida_aimall.

Register parsers via the "aiida.parsers" entry point in pyproject.toml.
"""

import numpy as np
import pandas as pd
from aiida.common import exceptions
from aiida.engine import ExitCode
from aiida.orm import Dict, SinglefileData
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory, DataFactory
from subproptools import qtaim_extract as qt  # pylint: disable=import-error

AimqbCalculation = CalculationFactory("aimall.aimqb")


class AimqbBaseParser(Parser):
    """
    Parser class for parsing output of calculation.
    """

    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced by a AimqbCalculation.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.nodes.process.process.ProcessNode`
        """
        super().__init__(node)
        if not issubclass(node.process_class, AimqbCalculation):
            raise exceptions.ParsingError("Can only parse AimqbCalculation")

    def parse(self, **kwargs):
        """Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """
        # convenience method to get filename of output file
        # output_filename = self.node.get_option("output_filename")
        input_parameters = self.node.inputs.parameters
        output_filename = self.node.process_class.OUTPUT_FILE

        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [
            output_filename.replace("out", "sum"),
            output_filename.replace(".out", "_atomicfiles"),
        ]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error(
                f"Found files '{files_retrieved}', expected to find '{files_expected}'"
            )
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        # parse output file
        self.logger.info(f"Parsing '{output_filename}'")
        OutFolderData = self.retrieved
        with OutFolderData.open(output_filename.replace("out", "sum"), "rb") as handle:
            output_node = SinglefileData(file=handle)
            sum_lines = output_node.get_content()
            out_dict = {
                "atomic_properties": self._parse_atomic_props(sum_lines),
                "bcp_properties": self._parse_bcp_props(sum_lines),
                "ldm": self._parse_ldm(sum_lines),
            }
        # if laprhocps were calculated, get cc_properties
        if "-atlaprhocps=True" in input_parameters.cmdline_params("foo"):
            out_dict["cc_properties"] = self._parse_cc_props(
                out_dict["atomic_properties"]
            )

        def make_serializeable(data):
            """Recursively go through the dictionary and convert unserializeable values in-place:

            1) In numpy arrays:
                * ``nan`` -> ``0.0``
                * ``inf`` -> large number

            :param data: A mapping of data.
            """
            if isinstance(data, dict):
                for key, value in data.items():
                    data[key] = make_serializeable(value)
            elif isinstance(data, list):
                for index, item in enumerate(data):
                    data[index] = make_serializeable(item)
            elif isinstance(data, np.ndarray):
                np.nan_to_num(data, copy=False)
            elif (
                not isinstance(data, dict)
                and not isinstance(data, np.ndarray)
                and not isinstance(data, list)
                and not isinstance(data, pd.DataFrame)
            ):
                if np.isnan(data):
                    data = np.nan_to_num(data)
            return data

        make_serializeable(out_dict)
        # store in node
        self.outputs.output_parameters = Dict(out_dict)

        return ExitCode(0)

    def _parse_ldm(self, sum_lines):
        return qt.get_ldm(sum_lines.split("\n"))

    def _parse_cc_props(self, atomic_properties):
        """Extract VSCC properties from output files
        :param atomic_properties: dictionary of atomic properties from _parse_atomic_props
        :param type atomic_properties: dict
        """
        output_filename = self.node.process_class.OUTPUT_FILE
        atom_list = list(atomic_properties.keys())
        # for each atom, load the .agpviz file in the _atomicfiles folder and get cc props
        cc_dict = {
            x: qt.get_atom_vscc(
                filename=self.retrieved.get_object_content(
                    output_filename.replace(".out", "_atomicfiles")
                    + "/"
                    + x.lower()
                    + ".agpviz"
                ).split("\n"),
                atomLabel=x,
                atomicProps=atomic_properties,
                is_lines_data=True,
            )
            for x in atom_list
        }
        return cc_dict

    def _parse_atomic_props(self, sum_file_string):
        """Extracts atomic properties from .sum file

        :param sum_file_string: lines of .sum output file
        :param type sum_file_string: str
        """
        return qt.get_atomic_props(sum_file_string.split("\n"))

    def _parse_bcp_props(self, sum_file_string):
        """Extracts bcp properties from .sum file

        :param sum_file_string: lines of .sum output file
        :param type sum_file_string: str
        """
        bcp_list = qt.find_all_connections(sum_file_string.split("\n"))
        return qt.get_selected_bcps(sum_file_string.split("\n"), bcp_list)


NUM_RE = r"[-+]?(?:[0-9]*[.])?[0-9]+(?:[eE][-+]?\d+)?"

SinglefileData = DataFactory("core.singlefile")


class AimqbGroupParser(AimqbBaseParser):
    """
    Parser class for parsing output of calculation.
    """

    def parse(self, **kwargs):
        """Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """
        # convenience method to get filename of output file
        # output_filename = self.node.get_option("output_filename")
        input_parameters = self.node.inputs.parameters
        output_filename = self.node.process_class.OUTPUT_FILE

        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [
            output_filename.replace("out", "sum"),
            output_filename.replace(".out", "_atomicfiles"),
        ]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error(
                f"Found files '{files_retrieved}', expected to find '{files_expected}'"
            )
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES
            # return

        # parse output file
        self.logger.info(f"Parsing '{output_filename}'")
        OutFolderData = self.retrieved
        with OutFolderData.open(output_filename.replace("out", "sum"), "rb") as handle:
            output_node = SinglefileData(file=handle)
            sum_lines = output_node.get_content()
            out_dict = {
                "atomic_properties": self._parse_atomic_props(sum_lines),
                "bcp_properties": self._parse_bcp_props(sum_lines),
                # "ldm": self._parse_ldm(sum_lines),
            }
        # if laprhocps were calculated, get cc_properties
        if "-atlaprhocps=True" in input_parameters.cmdline_params("foo"):
            out_dict["cc_properties"] = self._parse_cc_props(
                out_dict["atomic_properties"]
            )
        out_dict["graph_descriptor"] = self._parse_graph_descriptor(out_dict)
        # store in node
        if self.node.inputs.group_atoms.get_list():
            group_nums = self.node.inputs.group_atoms.get_list()

            out_dict["group_descriptor"] = self._parse_group_descriptor(
                out_dict["atomic_properties"], group_nums
            )
        else:  # default to using only atom # 2 as the substrate
            num_ats = len(out_dict["atomic_properties"])
            group_nums = [x + 1 for x in range(num_ats) if x != 1]
            out_dict["group_descriptor"] = self._parse_group_descriptor(
                out_dict["atomic_properties"], group_nums
            )
        self.outputs.output_parameters = Dict(out_dict)

        return ExitCode(0)

    def _parse_graph_descriptor(self, out_dict):
        """Get atomic, BCP, and VSCC properties of atom 1"""
        graph_dict = {}
        at_id = self.node.inputs.attached_atom_int.value
        # Find the atom property dictionary corresponding to the attached atom
        # Also add the atomic symbol to the property dictionary
        for key, value in out_dict["atomic_properties"].items():
            at_num = int("".join(x for x in key if x.isdigit()))
            if at_num == at_id:
                graph_dict["attached_atomic_props"] = value
                graph_dict["attached_atomic_props"]["symbol"] = "".join(
                    x for x in key if not x.isdigit()
                )
                break
        graph_dict["attached_bcp_props"] = {}
        for key, value in out_dict["bcp_properties"].items():
            num_bond = "".join(x for x in key if x.isdigit() or x == "-")
            at_nums = num_bond.split("-")
            if str(at_id) in at_nums:
                graph_dict["attached_bcp_props"][key] = value
        if "cc_properties" in list(out_dict.keys()):
            for key, value in out_dict["cc_properties"].items():
                at_num = int("".join(x for x in key if x.isdigit()))
                if at_num == at_id:
                    graph_dict["attached_cc_props"] = value
                    graph_dict["attached_cc_props"]["symbol"] = "".join(
                        x for x in key if not x.isdigit()
                    )
                    break
        return graph_dict

    def _parse_group_descriptor(self, atomic_properties, sub_atom_ints):
        """Convert atomic properties to group properties given atoms in group to use"""
        atoms = list(atomic_properties.keys())
        return qt.get_sub_props(atomic_properties, sub_atom_ints, atoms)



"""Workchains that are smaller parts of SubstituenParamWorkChain"""
# pylint:disable=no-member
from aiida.engine import ToContext, WorkChain, if_
from aiida.orm import (
    Bool,
    Code,
    Dict,
    Int,
    SinglefileData,
    Str,
    StructureData,
    load_group,
)
from aiida.orm.extras import EntityExtras
from aiida_gaussian.calculations import GaussianCalculation

from aiida_aimall.calculations import AimqbCalculation
from aiida_aimall.data import AimqbParameters
from aiida_aimall.workchains.calcfunctions import (
    create_wfx_from_retrieved,
    dict_to_structure,
    generate_rotated_structure_aiida,
    generate_structure_data,
    get_substituent_input,
    get_wfxname_from_gaussianinputs,
    parameters_with_cm,
)


class SmilesToGaussianWorkChain(WorkChain):
    r"""Workchain to take a substituent SMILES, and run a Gaussian calculation on that SMILES

    Takes an input SMILES with one placeholder \*, generates a geometry with \* replaced with a hydrogen.
    U

    Attributes:
        smiles (aiida.orm.Str): SMILES of a substiuent. Must contain a single placeholder \*
        gaussian_parameters (aiida.orm.Dict): Gaussian calculation for generating a wfx
        gaussian_code (aiida.orm.Code): Gaussian Code
        wfxname (aiida.orm.Str): name of wfx file provided in gaussian_parameters
        wfxgroup (aiida.orm.Str): group to store the wfx file in
        mem_mb (aiida.orm.Int): amount of memory in MB for the Gaussian calculation
        nprocs (aiida.orm.Int): number of processors for the Gaussian calculation
        time_s (aiida.orm.Int): amount of time to run the Gaussian calculation

    Note:
        The SMILES provided should have a single \*.

    Note:
        Uses the charge and multiplicity of the provided SMILES, not that provided to gaussian_parameters

    Note:
        'output':'wfx' should be provided to `gaussian_parameters`. And a .wfx file name should be provided
        as well

    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "smiles",
            valid_type=Str,
            help="SMILES of a substiuent. Must contain a single placeholder *",
        )
        spec.input(
            "gaussian_parameters",
            valid_type=Dict,
            help="Gaussian calculation for generating a wfx",
        )
        spec.input("gaussian_code", valid_type=Code, help="Gaussian Code")
        spec.input(
            "wfxname",
            required=False,
            valid_type=Str,
            help="name of wfx file provided in gaussian_parameters",
        )
        spec.input(
            "wfxgroup",
            required=False,
            valid_type=Str,
            help="group to store the wfx file in",
        )
        spec.input("nprocs", default=lambda: Int(4))
        spec.input("mem_mb", default=lambda: Int(6400))
        spec.input("time_s", default=lambda: Int(24 * 7 * 60 * 60))
        spec.input("dry_run", default=lambda: Bool(False))
        spec.output("wfx", valid_type=SinglefileData, required=False)
        spec.output("output_parameters", valid_type=Dict)
        spec.outline(
            cls.get_substituent_inputs_step,  # , cls.results
            cls.update_parameters_with_cm,
            cls.string_to_StructureData,
            cls.get_wfx_name,
            cls.submit_gaussian,
            if_(cls.found_wfx_name)(cls.create_wfx_file),
            cls.results,
        )

    def get_substituent_inputs_step(self):
        """Given list of substituents and previously done smiles, get input"""
        self.ctx.smiles_geom = get_substituent_input(self.inputs.smiles)

    def update_parameters_with_cm(self):
        """Update provided Gaussian parameters with charge and multiplicity of substituent"""
        self.ctx.gaussian_cm_params = parameters_with_cm(
            self.inputs.gaussian_parameters, self.ctx.smiles_geom
        )

    def string_to_StructureData(self):
        """Convert an xyz string of molecule geometry to StructureData"""
        self.ctx.structure = generate_structure_data(self.ctx.smiles_geom)

    def get_wfx_name(self):
        """Find the wavefunction file in the retrieved node"""
        if "wfxname" in self.inputs:
            self.ctx.wfxname = self.inputs.wfxname
        else:
            self.ctx.wfxname = get_wfxname_from_gaussianinputs(
                self.inputs.gaussian_parameters
            )

    def submit_gaussian(self):
        """Submits the gaussian calculation"""
        builder = GaussianCalculation.get_builder()

        builder.structure = self.ctx.structure
        builder.parameters = self.ctx.gaussian_cm_params
        builder.code = self.inputs.gaussian_code
        builder.metadata.options.resources = {
            "num_machines": 1,
            "tot_num_mpiprocs": self.inputs.nprocs.value,
        }
        builder.metadata.options.max_memory_kb = (
            int(self.inputs.mem_mb.value * 1.25) * 1024
        )
        builder.metadata.options.max_wallclock_seconds = self.inputs.time_s.value
        if self.ctx.wfxname.value:
            builder.metadata.options.additional_retrieve_list = [
                self.ctx.wfxname.value.strip()
            ]

        if self.inputs.dry_run.value:
            return self.inputs
        node = self.submit(builder)
        out_dict = {"opt": node}
        return ToContext(out_dict)

    def found_wfx_name(self):
        """Check if we found a wfx or wfn file"""
        if self.ctx.wfxname.value:
            return True
        return False

    def create_wfx_file(self):
        """Create a wavefunction file from the retireved folder"""
        retrieved_folder = (
            self.ctx["opt"].base.links.get_outgoing().get_node_by_label("retrieved")
        )
        wfx_node = create_wfx_from_retrieved(self.ctx.wfxname, retrieved_folder)
        wfx_node.base.extras.set("smiles", self.inputs.smiles)
        self.ctx.wfxfile = wfx_node
        if "wfxgroup" in self.inputs:
            wfx_group = load_group(self.inputs.wfxgroup.value)
            wfx_group.add_nodes(wfx_node)

    def results(self):
        """Store our relevant information as output"""
        if "wfxfile" in self.ctx:
            self.out("wfx", self.ctx.wfxfile)
        self.out(
            "output_parameters",
            self.ctx["opt"]
            .base.links.get_outgoing()
            .get_node_by_label("output_parameters"),
        )


class AIMAllReorWorkChain(WorkChain):
    """Workchain to run AIM and then reorient the molecule using the results

    Often called in `aiida_aimall.controllers.AimReorSubmissionController`.
    Process continues in `aiida_aimall.controllers.GaussianSubmissionController`.

    Attributes:
        aim_params: (AimqbParameters): Command line parameters for aimqb
        file (aiida.orm.SinglefileData): .fchk, .wfn, or .wfx file for aimqb input
        aim_code (aiida.orm.Code): AIMQB code
        frag_label (aiida.orm.Str): Optional SMILES tag of the substituent
        aim_group (aiida.orm.Str): Optional group to put the AIM calculation node in
        reor_group (aiida.orm.Str): Optional group to put the reoriented structure in

    Example

        ::

            from aiida_aimall.data import AimqbParameters
            from aiida_aimall.workchains.param_parts import AIMAllReorWorkChain
            from aiida.orm import SinglefileData, load_code
            from aiida.engine import submit
            input_file = SinglefileData("/absolute/path/to/file")
            aim_code = load_code("aimall@localhost")
            aim_params = AimqbParameters({'nproc':2,'naat':2,'atlaprhocps':True})
            builder = AIMAllReorWorkChain.get_builder()
            builder.file = input_file
            builder.aim_code = aim_code
            builder.aim_params = aim_params
            submit(builder)

    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "aim_params",
            valid_type=AimqbParameters,
            help="Command line parameters for aimqb",
        )
        spec.input(
            "file",
            valid_type=SinglefileData,
            help="fchk, wfn, or wfx file for aimqb input",
        )
        # spec.output('aim_dict',valid_type=Dict)
        spec.input("aim_code", valid_type=Code, help="aimqb code")
        spec.input(
            "frag_label",
            valid_type=Str,
            required=False,
            help="Optional SMILES tag of the substituent",
        )
        spec.input(
            "aim_group",
            valid_type=Str,
            required=False,
            help="Optional group to put the AIM calculation node in",
        )
        spec.input(
            "reor_group",
            valid_type=Str,
            required=False,
            help="Optional group to put the reoriented structure in",
        )
        spec.input("dry_run", valid_type=Bool, default=lambda: Bool(False))
        spec.output("rotated_structure", valid_type=StructureData)
        spec.outline(cls.aimall, cls.rotate, cls.dict_to_struct_reor, cls.result)

    def aimall(self):
        """submit the aimall calculation"""
        builder = AimqbCalculation.get_builder()
        builder.code = self.inputs.aim_code
        builder.parameters = self.inputs.aim_params
        builder.file = self.inputs.file
        builder.metadata.options.resources = {
            "num_machines": 1,
            "tot_num_mpiprocs": 2,
        }
        if self.inputs.dry_run.value:
            return self.inputs
        aim_calc = self.submit(builder)
        aim_calc.store()
        if "aim_group" in self.inputs:
            aim_noreor_group = load_group(self.inputs.aim_group)
            aim_noreor_group.add_nodes(aim_calc)
        out_dict = {"aim": aim_calc}
        return ToContext(out_dict)

    def rotate(self):
        """perform the rotation"""
        aimfolder = (
            self.ctx["aim"].base.links.get_outgoing().get_node_by_label("retrieved")
        )
        output_dict = (
            self.ctx["aim"]
            .base.links.get_outgoing()
            .get_node_by_label("output_parameters")
            .get_dict()
        )
        atom_props = output_dict["atomic_properties"]
        cc_props = output_dict["cc_properties"]
        self.ctx.rot_struct_dict = generate_rotated_structure_aiida(
            aimfolder, atom_props, cc_props
        )

    def dict_to_struct_reor(self):
        """generate the gaussian input from rotated structure"""
        structure = dict_to_structure(self.ctx.rot_struct_dict)
        structure.store()
        if "reor_group" in self.inputs:
            reor_struct_group = load_group(self.inputs.reor_group.value)
            reor_struct_group.add_nodes(structure)
        if "frag_label" in self.inputs:
            struct_extras = EntityExtras(structure)
            struct_extras.set("smiles", self.inputs.frag_label.value)
        self.ctx.rot_structure = structure

    def result(self):
        """Parse results"""
        self.out("rotated_structure", self.ctx.rot_structure)



"""Base input workchain"""
from aiida.engine import WorkChain, if_
from aiida.engine.processes import ExitCode
from aiida.orm import SinglefileData, Str, StructureData

from aiida_aimall.workchains.calcfunctions import (
    generate_structure_data,
    get_molecule_str_from_smiles,
    xyzfile_to_StructureData,
)


class BaseInputWorkChain(WorkChain):
    """A workchain to generate and validate inputs.

    Provided an .xyz file as `SinglefileData`, molecule `StructureData`, or SMILES of the molecule
    validates that only one is provided. Then, prepares the input into a format for future GaussianCalculations.

    Attributes:
        structure (aiida.orm.StructureData): StructureData of molecule to run
        smiles (aiida.orm.Str): smiles string of molecule
        xyz_file (aiida.orm.SinglefileData): file data of an xyz file

    Note:
        This is a base class that is used by other WorkChains
            (:func:`aiida_aimall.workchains.subparam.SubstituentParameterWorkChain`, and
            :func:`aiida_aimall.workchains.qc_programs.GaussianToAIMWorkChain`)
    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "structure",
            valid_type=StructureData,
            required=False,
            help="StructureData of molecule to run",
        )
        spec.input(
            "smiles", valid_type=Str, required=False, help="smiles string of molecule"
        )
        spec.input(
            "xyz_file",
            valid_type=SinglefileData,
            required=False,
            help="file data of an xyz file",
        )
        spec.exit_code(
            200,
            "ERROR_MULTIPLE_INPUTS",
            "the process received two or more of the following inputs: structure, smiles, xyz_file",
        )
        spec.exit_code(
            201,
            "ERROR_NO_INPUTS",
            "None of structure, smiles, xyz_file were provided, at least one must be",
        )
        spec.outline(
            cls.validate_input,
            if_(cls.is_xyz_input)(cls.create_structure_from_xyz),
            if_(cls.is_smiles_input)(
                cls.get_molecule_inputs_step, cls.string_to_StructureData
            ),
            if_(cls.is_structure_input)(cls.structure_in_context),
        )

    def is_xyz_input(self):
        """Validates if xyz_file was provided as input"""
        if "xyz_file" in self.inputs:
            return True
        return False

    def is_smiles_input(self):
        """Validates if smiles was provided as input"""
        if "smiles" in self.inputs:
            return True
        return False

    def is_structure_input(self):
        """Validates if structure was provided as input"""
        if "structure" in self.inputs:
            return True
        return False

    # pylint:disable=inconsistent-return-statements
    def validate_input(self):
        """Check that only one of smiles, structure, or xyz_file was input"""
        if "smiles" in self.inputs and (
            "xyz_file" in self.inputs or "structure" in self.inputs
        ):
            return ExitCode(200)
        if "xyz_file" in self.inputs and (
            "smiles" in self.inputs or "structure" in self.inputs
        ):
            return ExitCode(200)
        if "structure" in self.inputs and (
            "xyz_file" in self.inputs or "smiles" in self.inputs
        ):
            return ExitCode(200)
        if (
            "structure" not in self.inputs
            and "xyz_file" not in self.inputs
            and "smiles" not in self.inputs
        ):
            return ExitCode(201)

    def create_structure_from_xyz(self):
        """Convert the xyzfile to StructureData. Calls
        :func:`aiida_aimall.workchains.calcfunctions.xyzfile_to_StructureData`"""
        self.ctx.structure = xyzfile_to_StructureData(self.inputs.xyz_file)

    def structure_in_context(self):
        """Store the input structure in context, to make consistent with the results of xyz_file or SMILES input."""
        self.ctx.structure = self.inputs.structure

    def get_molecule_inputs_step(self):
        """Given list of substituents and previously done smiles, get input.
        Calls :func:`aiida_aimall.workchains.calcfunctions.get_molecule_str_from_smiles`"""
        self.ctx.smiles_geom = get_molecule_str_from_smiles(self.inputs.smiles)

    def string_to_StructureData(self):
        """Convert an xyz string of molecule geometry to StructureData.
        Calls :func:`aiida_aimall.workchains.calcfunctions.generate_structure_data`"""
        self.ctx.structure = generate_structure_data(self.ctx.smiles_geom)



"""`CalcJob` implementation for the aimqb executable of AIMAll."""
from aiida.common import datastructures
from aiida.engine import CalcJob
from aiida.orm import Dict, Int, List, SinglefileData

from aiida_aimall.data import AimqbParameters


class AimqbCalculation(CalcJob):
    """AiiDA calculation plugin wrapping the aimqb executable.

    Attributes:
        parameters (AimqbParameters): command line parameters for the AimqbCalculation
        file (aiida.orm.SinglefileData): the wfx, wfn, or fchk file to be run
        code (aiida.orm.Code): code of the AIMQB executable
        attached_atom_int (aiida.orm.Int): optional integer label of the atom that is attached to the rest of the molecule
        group_atoms (aiida.orm.List): optional integer list of ids of atoms comprising the group for AimqbGroupParser

    Example:
        ::

            code = orm.load_code('aimall@localhost')
            AimqbParameters = DataFactory("aimall.aimqb")
            aim_params = AimqbParameters(parameter_dict={"naat": 2, "nproc": 2, "atlaprhocps": True})
            file = SinglefileData("/absolute/path/to/file")
            # Alternatively, if you have the file as a string, you can build the file with:
            # file=SinglefileData(io.BytesIO(file_string.encode()))
            AimqbCalculation = CalculationFactory("aimall.aimqb")
            builder  = AimqbCalculation.get_builder()
            builder.parameters = aim_params
            builder.file = file
            builder.code = code
            builder.metadata.options.resources = {"num_machines": 1, "num_mpiprocs_per_machine": 2}
            builder.submit()

    Note:
        By default, the AimqbBaseParser is used, getting atomic, BCP, and (if applicable) LapRhoCps.
            You can opt to use the AimqbGroupParser, which also returns the integrated group properties
            of a group, as well as the atomic graph descriptor of the group. In doing so, you can also
            define the atoms included in the group, which, by convention, defaults to all atoms except atom 2.
            You can further specify which atom of the group is the one bonded to the substrate, which defaults to
            atom 1.  This is done by providing this to the builder:

        ::

            builder.metadata.options.parser_name = "aimall.group"
            builder.attached_atom_int = Int(1)
            builder.group_atoms = List([1,3,4,5,6])

    """

    INPUT_FILE = "aiida.wfx"
    OUTPUT_FILE = "aiida.out"
    PARENT_FOLDER_NAME = "parent_calc"
    DEFAULT_PARSER = "aimall.base"

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation"""
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs["metadata"]["options"]["resources"].defaults = {
            "num_machines": 1,
            "tot_num_mpiprocs": 2,
        }

        spec.inputs["metadata"]["options"]["parser_name"].default = "aimall.base"

        spec.input(
            "attached_atom_int",
            valid_type=Int,
            help="id # of attached atom for graph descriptor. Defaults to atom 1",
            default=lambda: Int(1),
        )
        spec.input(
            "group_atoms",
            valid_type=List,
            help="Integer ids of atoms in groups to include. e.g. [1,3,4]. Defaults to all atoms in molecule except atom 2",
            default=lambda: List([]),
        )
        spec.input(
            "parameters",
            valid_type=AimqbParameters,
            help="Command line parameters for aimqb",
        )
        spec.input(
            "file",
            valid_type=SinglefileData,
            help="fchk, wfn, or wfx to run AimQB on",
        )

        spec.output(
            "output_parameters",
            valid_type=Dict,
            required=True,
            help="The computed parameters of an AIMAll calculation",
        )

        spec.default_output_node = "output_parameters"
        spec.exit_code(
            210,
            "ERROR_MISSING_OUTPUT_FILES",
            message="The retrieved folder did not contain the output file.",
        )
        spec.outputs.dynamic = True

    def prepare_for_submission(self, folder):
        """Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily
            place all files needed by the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        # copy wfx file to input file
        input_string = self.inputs.file.get_content()
        with open(
            folder.get_abs_path(self.INPUT_FILE), "w", encoding="utf-8"
        ) as out_file:
            out_file.write(input_string)
        codeinfo = datastructures.CodeInfo()
        # generate command line params
        codeinfo.cmdline_params = self.inputs.parameters.cmdline_params(
            file_name=self.INPUT_FILE
        )
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.OUTPUT_FILE

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]  # list since can involve more than one
        # Retrieve the sum file and the folder with atomic files
        calcinfo.retrieve_list = [
            self.OUTPUT_FILE.replace("out", "sum"),
            self.OUTPUT_FILE.replace(".out", "_atomicfiles"),
        ]

        return calcinfo
